from doctest import OutputChecker
import glob
import os

import click
from omegaconf import OmegaConf
from scabha.basetypes import File
from scabha.schema_utils import clickify_parameters, paramfile_loader
import zarr

import visco
from visco import BIN, get_logger
from visco import decompress

log = get_logger(BIN.decompression)

command = BIN.decompression

thisdir = os.path.dirname(__file__)
decompression_params = glob.glob(f"{thisdir}/*.yaml")
decompression_files = [File(item) for item in decompression_params]
parserfile = File(f"{thisdir}/{command}.yaml")
config = paramfile_loader(parserfile, decompression_files)[command]


@click.command(command)
@click.version_option(str(visco.__version__))
@clickify_parameters(config)
def runit(**kwargs):
    opts = OmegaConf.create(kwargs)
    zarr_path = opts.zarr_path
    output_column = opts.output_column
    ms = opts.ms
    
    decompress.decompress_visdata(zarr_path, output_column, ms)


from omegaconf import OmegaConf
import zarr
import numpy as np
import dask, daskms
import numpy as np
from daskms import xds_to_table
import xarray as xr
import visco
import dask.array as da
from casacore.tables import table
log = visco.get_logger(name="VISCO")
import logging
from tqdm.dask import TqdmCallback
logging.getLogger("daskms").setLevel(logging.ERROR)

#The correlation mapping.
CORR_TYPES = OmegaConf.load(f"{visco.PCKGDIR}/ms_corr_types.yaml").CORR_TYPES


def reconstruction(U,S,WT):
    """"
    Reconstruct the compressed data using the components U,S,WT.
    
    Params
    ----
    U: array
    - The left singular matrix.
    
    S: array
    - The singular values matrix.
    
    WT:array
    - The right singular matrix
    
    Returns
    ----
    The reconstructed data.
    """
    
    recon_data = U @ da.diag(S) @ WT
    
    return recon_data

def decompress_visdata(zarr_path, output_column='DATA',output_ms='decompressed.ms'):
    """
    Decompress visibility data from a Zarr file.
    
    Params
    ----
    zarr_path: str
    - The full path to the zarr file containing the compressed data.
    
    output_column:str
    - The column to put the visibility data in the MS.
    
    output_ms:str
    - The name of the MS to be created.
    
    Return
    ----
    An MS with the reconstructed data.
    
    """
    
    #Maintable and subtables of the Zarr file.
    maintable = xr.open_zarr(zarr_path)
   
    zarr_open = zarr.open(zarr_path, mode='r')
    spw = xr.open_zarr(zarr_path,group='SPECTRAL_WINDOW')
    fld = xr.open_zarr(zarr_path,group='FIELD')
    pol = xr.open_zarr(zarr_path,group='POLARIZATION')
    pointing = xr.open_zarr(zarr_path,group='POINTING')
    antenna = xr.open_zarr(zarr_path,group='ANTENNA')
    flag_row_zarr = xr.open_zarr(zarr_path,group='FLAG_ROW')
    flag_zarr = xr.open_zarr(zarr_path,group='FLAG')
    
    #Shape and chunk of the data.
    shape = maintable.attrs['shape']
    chunks = maintable.chunksizes['row'][0]
   
    #Initialize zeros.
    decompressed_data = da.zeros(shape=shape,dtype=complex,chunks=chunks)
    
    data_group_zarr = zarr_open['DATA']
    baseline_keys = list(data_group_zarr.group_keys())
    
    #Go through all the baselines and correlations.
    for baseline_key in baseline_keys:
        baseline_x = data_group_zarr[baseline_key]
        corr_keys = list(baseline_x.group_keys())
    
        for ci,corr_key in enumerate(corr_keys):
            group_path = f"DATA/{baseline_key}/{corr_key}"
            corr_group = xr.open_zarr(zarr_path, group=group_path)
            
            U = corr_group.U.data
            S = corr_group.S.data
            WT = corr_group.WT.data
            baseline_filter = corr_group.attrs['baseline_filter']
            
            reconstructed_data = reconstruction(U, S, WT)
            
            decompressed_data[baseline_filter, :, ci] = reconstructed_data
     
    nrow =  decompressed_data.shape[0] 
    nchan = decompressed_data.shape[1]
    ncorr = decompressed_data.shape[2]
    
    #Write the reconstructed data to maintable dataset.
    maintable = maintable.assign(**{
    output_column: xr.DataArray(decompressed_data, 
                                dims=("row", "chan", "corr"),
                                coords={
                                    "row": np.arange(nrow),
                                    "chan": np.arange(nchan),
                                    "corr": np.arange(ncorr)
                                })
})

    #Decompress the flag data and write it to the maintable dataset.         
    flag_row_data = flag_row_zarr.FLAG_ROW.values
    flag_row_data_unpacked = decompress_bits(flag_row_data,dim=1)  
    
    maintable = maintable.assign(**{
    "FLAG_ROW": xr.DataArray(da.from_array(flag_row_data_unpacked,chunks=chunks), 
                                dims=("row"),
                                coords={
                                    "row": np.arange(nrow),
                                }) 
    })      
    
    
    flag_data = flag_zarr.FLAG.values
    flag_data_unpacked = decompress_bits(flag_data,dim=3,nrow=nrow,nchan=nchan,ncorr=ncorr)  
    
    maintable = maintable.assign(**{
    "FLAG": xr.DataArray(da.from_array(flag_data_unpacked,chunks=chunks), 
                                dims=("row", "chan", "corr"),
                                coords={
                                    "row": np.arange(nrow),
                                    "chan": np.arange(nchan),
                                    "corr": np.arange(ncorr)
                                }) 
    })      
    
    
    #Finally, write the MS.
    write_main = xds_to_table(maintable, output_ms)
    with TqdmCallback(desc=f'Writing Main Table to {output_ms}'):
        
        dask.compute(write_main)   
        
    write_spw = xds_to_table(spw, f"{output_ms}::SPECTRAL_WINDOW")
    with TqdmCallback(desc=f'Writing SPECTRAL_WINDOW table to {output_ms}'):
        
        dask.compute(write_spw)
        
        
    write_fld = xds_to_table(fld, f"{output_ms}::FIELD")
    with TqdmCallback(desc=f'Writing FIELD table to {output_ms}'):
        
        dask.compute(write_fld)
        
    write_pol = xds_to_table(pol, f"{output_ms}::POLARIZATION")
    with TqdmCallback(desc=f'Writing POLARIZATION table to {output_ms}'):
        
        dask.compute(write_pol)
    
    write_antenna = xds_to_table(antenna, f"{output_ms}::ANTENNA")
    with TqdmCallback(desc=f'Writing ANTENNA table to {output_ms}'):
        
        dask.compute(write_antenna)
        
    write_pointing = xds_to_table(pointing, f"{output_ms}::POINTING")
    with TqdmCallback(desc=f'Writing POINTING table to {output_ms}'):
        
        dask.compute(write_pointing)
        
  
    
def decompress_bits(data_array,dim=1, nrow=None,nchan=None,ncorr=None):
    """"
    Decompress data (flag) using numpy unpackbits.
    
    Params
    ----
    data_array: array
    - The data array to be compressed.
    
    dim: int
    - The number of dimensions of the data.
    
    nrow: int
    - The number of rows of the visibility data.
    
    nchan: int
    - The number of channels of the visibility data.
    
    ncorr: int
    - THe number of correlations of the visibility data.
    """
    
    #If the data has one dimension, its easy. Just make sure we account for padding by 
    #selecting the first nrow.
    if dim==1:
        unpacked_data = np.unpackbits(data_array,axis=None)
        unpacked_data = unpacked_data[:nrow]
        
    #If the data has more than one dimension, we have to reshape the data to the original because
    #it has been flattened.
    elif dim>1:
        undone_data = np.unpackbits(data_array,axis=None)
        undone_data = undone_data[:nrow*nchan*ncorr]
        unpacked_data = undone_data.reshape(nrow,nchan,ncorr)
    return unpacked_data    

def ms_addrow(ms,subtable,nrows):
    
    subtab = table(f"{ms}::{subtable}",
                    readonly=False, lockoptions='user', ack=False)
    try:
        subtab.lock(write=True)
        subtab.addrows(nrows)
        
    finally:
        subtab.unlock()
        subtab.close()
        
       
def ms_remrow(ms,subtable,row):
    
    subtab = table(f"{ms}::{subtable}",
                    readonly=False, lockoptions='user', ack=False)
    try:
        subtab.lock(write=True)
        subtab.removerows(row)
        
    finally:
        subtab.unlock()
        subtab.close()


import glob
import os

import click
from omegaconf import OmegaConf
from scabha.basetypes import File
from scabha.schema_utils import clickify_parameters, paramfile_loader

import visco
from visco import BIN, get_logger
from visco import archive

log = get_logger(BIN.archival)

command = BIN.archival

thisdir = os.path.dirname(__file__)
archival_params = glob.glob(f"{thisdir}/*.yaml")
archival_files = [File(item) for item in archival_params]
parserfile = File(f"{thisdir}/{command}.yaml")
config = paramfile_loader(parserfile, archival_files)[command]



@click.command(command)
@click.version_option(str(visco.__version__))
@clickify_parameters(config)
def runit(**kwargs):
    opts = OmegaConf.create(kwargs)
    ms = opts.ms
    fieldid = opts.fieldid
    ddid = opts.ddid
    scan = opts.scan
    column = opts.column
    autocorrelation = opts.autocorrelation
    antlist = opts.antlist
    flagvalue = opts.flagvalue
    correlation = opts.correlation
    outfilename = opts.outfilename
    compressionrank = opts.compressionrank
    decorrelation = opts.decorrelation
    archive.archive_visdata(ms, correlation, fieldid, ddid, scan,
                      column, outfilename, compressionrank,
                      autocorrelation, decorrelation, antlist, flagvalue)
    


import dask.array as da
import xarray as xr
import os
import ast
from itertools import combinations
from daskms import xds_from_table
import numpy as np
from copy import copy
from tqdm.dask import TqdmCallback
from visco.utilities import ObjDict
import visco
log = visco.get_logger(name="VISCO")
import logging
logging.getLogger("daskms").setLevel(logging.ERROR)

from omegaconf import OmegaConf


def openms(ms,sub=None,query=None):
    """
        Open a Meaurement Set.
        
        Params
        ----
        ms: str 
        - The path to the MS.
        
        sub: str 
        - The name of the subtable to be opened.
        
        query: str 
        - A query to send to the MS.
        
        Returns
        ----
        
        An xarray dataset.
        
    """
    #If subtable is given
    if sub:
        #If query is given
        if query:
            ds = xds_from_table(f"{ms}::{sub}",taql_where=query)
        else:
            ds = xds_from_table(f"{ms}::{sub}")
    else:
        #If opening the main table
        if query:
            ds = xds_from_table(ms,taql_where=query)
        else:
            ds = xds_from_table(ms)
     
    #Accessing all the chunks in the dataset   
    for dsi in ds:
        dataset = dsi
    return dataset
        
      
def decompose(ms,correlation, fieldid,ddid,scan,
                column,autocorrelation,antlist,flagvalue,maintab,fieldtab,anttab,poltab):
    

    """
    Decompose the visibility data matrix for each baseline and 
    store into a dictionary.
    
    Params
    ----
    
    ms: str
    - The path to the MS.
    
    correlation: str
    - The correlation to decompose.
    
    fieldid: int
    - The FIELD_ID to use for the decomposition.
    
    ddid: int
    - The DATA_DESC_ID to use for the decomposition.
    
    scan: int
    - The SCAN_NUMBER to use.
    
    column: str
    - The data column to decompose.
    
    autocorrelation: bool
    - Whether there is autocorrelation or not.
    
    antlist: List[int]
    - The antenna indices to use.
    
    flagvalue: int or float
    - The value to replace the data that is flagged.
    
    maintab: xarray dataset
    - The main table of the measurement set.
    
    fieldtab: xarray dataset
    - The FIELD table of the measurement set.
    
    anttab: xarray dataset
    - The ANTENNA table of the measurement set.
    
    poltab: xarray dataset
    - The POLARIZATION table of the measurement set.
    
    Returns 
    ---
    visdata_dict: dict
    - A dictionary with the decomposed data.
    
    """
    CORR_TYPES = OmegaConf.load(f"{visco.PCKGDIR}/ms_corr_types.yaml").CORR_TYPES
    CORR_TYPES_REVERSE = OmegaConf.load(f"{visco.PCKGDIR}/ms_corr_types_reverse.yaml").CORR_TYPES
    
    #Open the measurement set and its subtables
    maintable = maintab
    fieldtable = fieldtab
    antennatable = anttab
    poltable = poltab
    
    #Get the needed data
    scan_number = maintable.SCAN_NUMBER.data
    data_desc_id = maintable.DATA_DESC_ID.data
    field_id = maintable.FIELD_ID.data
    ant1 = maintable.ANTENNA1.data.compute()
    ant2 = maintable.ANTENNA2.data.compute()
    corr_types = poltable.CORR_TYPE.data.compute()
    field_names = fieldtable.NAME.data
    
    #Create a mask 
    fidmsk = (field_id.compute() == fieldid)
    didmsk = (data_desc_id.compute() == ddid)
    scnmsk = (scan_number.compute() == scan)
    
    #Check if the given scan number, ddid, and fieldid exists.
    if scan not in scan_number:
        available_scans = da.unique(scan_number).compute().tolist()
        raise ValueError(f"Invalid SCAN_NUMBER {scan}. Available scans are: {available_scans}")
        
    
    if ddid not in data_desc_id:
        raise ValueError(f"Invalid selected DATA_DESC_ID {ddid}.\
                            Available DATA_DESC_IDs are {da.unique(data_desc_id)}")
    
    
    if fieldid < 0 or fieldid >= field_names.shape[0]:
        raise ValueError(f"Invalid selected FIELD_ID {fieldid}. There are\
                            {field_names.shape[0]} fields.")
    
    log.info(f"Decomposing visibilities for DATA_DESC_ID {ddid}. FIELD_ID {fieldid}, and SCAN_NUMBER {scan}.")
     
    #If given the antenna indices, use them to get baselines. 
    if antlist:
        
        if isinstance(antlist, str): 
            try:
                antlist = ast.literal_eval(antlist)
                if not isinstance(antlist, list) or not all(isinstance(x, int) for x in antlist):
                    raise ValueError("Parsed antlist is not a valid list of integers.")
            except (ValueError, SyntaxError):
                raise ValueError(f"Invalid format for antlist: {antlist}. Expected a list of integers.")
        
        baselines = list(combinations(antlist, 2))
        nbaselines = len(baselines)
             
    else:
        #If there are autocorrelations, which probably will never happen, consider it.
        if autocorrelation:
            baselines = np.unique(list(zip(ant1,ant2)),axis=0)
        else:
            baselines = np.unique([(a1, a2) for a1, a2 in zip(ant1,ant2) if a1 != a2],axis=0)
        nbaselines = baselines.shape[0]
    
    corr_list = []
    for corr in corr_types:
        for cor in corr:
            corr_name = CORR_TYPES_REVERSE[int(cor)]
            corr_list.append(corr_name)
    log.info(f"The following correlations are available:{corr_list}")
    
    corr_list_user = []
    for corr in correlation.split(','):
        corr_ind = CORR_TYPES[str(corr)]
        corr_list_user.append(corr_ind)
    log.info(f"The user has selected the following correlations:{list(correlation.split(','))}") 
    
    #Initialize the data dictionary to contain the decomposed components.
    vis_dict = {}
    
    #Go through all the baselines.
    for bx,(antenna1,antenna2) in enumerate(baselines):
        
        vis_dict[bx] = {}
        
        ant1msk = (ant1 == antenna1)
        ant2msk = (ant2 == antenna2)
        baseline_filter = fidmsk & didmsk & scnmsk & ant1msk & ant2msk
        
        uvw = maintable.UVW.data[baseline_filter]
        baseline_length = da.sqrt(uvw[:,0]**2 + uvw[:,1]**2)
        vis_dict[bx]["length"] = baseline_length
        vis_dict[bx]["baseline_filter"] = baseline_filter
        
        
        ant1name = antennatable.NAME.values[antenna1]
        ant2name = antennatable.NAME.values[antenna2]
    
        vis_dict[bx]["ant1name"] = ant1name
        vis_dict[bx]["ant2name"] = ant2name
        
        #Go through the given correlations
        for c in corr_list_user:
            ci = np.where(corr_types == c)[0][0]
            flag = maintable.FLAG.data[baseline_filter,:,ci]
            vis_data = copy(maintable[column].data[baseline_filter,:,ci])
            vis_data[flag] = flagvalue
            U,singvals,WT = da.linalg.svd(vis_data)
            fullrank = min(vis_data.shape[0],vis_data.shape[1])
    
        
            corr_type = CORR_TYPES_REVERSE[c]
            
            
            vis_dict[bx][corr_type] = ObjDict({
                "data": (U, singvals, WT),
                "shape": vis_data.shape,
                "corr": corr_type,
            
            })
            
            
            log.info(f"Decomposing visibility data for baseline {ant1name}-{ant2name}"
                     f" and correlation {corr_type}")
    
    return vis_dict

            



def archive_visdata(ms, correlation='XX,XY,YX,YY', fieldid=0, ddid=0, scan=1,
                column='DATA', outfilename='compressed-data.zarr', compressionrank=None,
                autocorrelation=False, decorrelation=None,
                antlist=None, flagvalue=0):
    
    """
    Compress the visibility data and store the decomposition components
    in Zarr file.
    
    Parameters
    ----
    
    ms: str
    - The path to the MS.
    
    correlation: str
    - The correlation to decompose.
    - default is 'XX,XY,YX,YY'.
    
    fieldid: int
    - The FIELD_ID to use for the decomposition.
    - Default is 0.
    
    ddid: int
    - The DATA_DESC_ID to use for the decomposition.
    - Default is 0.
    
    scan: int
    - The SCAN_NUMBER to use.
    - Default is 1.
    
    column: str
    - The data column to decompose.
    - Default is DATA.
    
    autocorrelation: bool
    - Whether there is autocorrelation or not.
    - Default is False.
    
    antlist: List[int]
    - The antenna indices to use.
    
    flagvalue: int or float
    - The value to replace the data that is flagged.
    - Default is 0.
    
    outfilename: str
    - The name of the output zarr file.
    - Default is compressed-data.zarr.
    
    compressionrank: int
    - The compression rank to apply uniformly on all the baselines.
    
    decorrelation: float
    - The minimum signal loss percentage on each baseline.
    
    Returns
    ----
    A zarr file with the data.
    """
    
    # Only one should be provided. Cannot provide both decorrelation and compressionrank
    if compressionrank is not None and decorrelation is not None:
        raise ValueError("Only one of 'compressionrank' or 'decorrelation' should be provided, not both.")
    elif compressionrank is None and decorrelation is None:
        raise ValueError("Either 'compressionrank' or 'decorrelation' must be provided.")
    
    #Open the tables
    maintable = openms(ms,query = f"FIELD_ID={fieldid} AND DATA_DESC_ID={ddid} AND SCAN_NUMBER={scan}")
    mtab = openms(ms)
    spw_table = openms(ms,"SPECTRAL_WINDOW")
    ant_table = openms(ms,"ANTENNA")
    pol_table = openms(ms,"POLARIZATION")
    fld_table = openms(ms,"FIELD")
    pntng_table = openms(ms,"POINTING")
    pointing_chunks = pntng_table.chunks["row"][0]
    data_row_chunks = maintable.chunks["row"][0]
    
    compressed_data = copy(maintable[column].data)
    
    #Full rank of the data.
    fullrank = min(compressed_data.shape[0],compressed_data.shape[1])
    log.info(f"The full rank of the data is {fullrank}.")
    
    #Get the decompositions.
    vis_data = decompose(ms=ms, correlation=correlation, fieldid=fieldid,
                         ddid=ddid, scan=scan, column=column,
                       autocorrelation=autocorrelation,antlist=antlist, flagvalue=flagvalue, 
                       maintab=mtab,fieldtab=fld_table,anttab=ant_table,poltab=pol_table)
    
    
    #The path to the Zarr file containing the compression results
    zarr_output_path = os.path.join(os.getcwd(), "zarr-output", f"{outfilename}")
    
    #Store the main table in the zarr file.
    ds = xr.Dataset(
    {
        "ANTENNA1": (("row"), maintable.ANTENNA1.values),
        "ANTENNA2": (("row"), maintable.ANTENNA2.values),
        "TIME": (("row"), maintable.TIME.values),
        "UVW": (("row", "uvw_dim"), maintable.UVW.values),
        "EXPOSURE": (("row"), maintable.EXPOSURE.values),
        "INTERVAL": (("row"), maintable.INTERVAL.values),
        "TIME_CENTROID": (("row"), maintable.TIME_CENTROID.values),
        "SCAN_NUMBER": (("row"), maintable.SCAN_NUMBER.values),
        "FIELD_ID": (("row"), maintable.FIELD_ID.values),
    },
    coords={
        "row": np.arange(len(maintable.ANTENNA1)),  
    },
    attrs={
        "shape": compressed_data.shape,
        "chunks": compressed_data.chunks,
        "ROWID": maintable.ROWID.values,
        },
    )
    
    #Writing to the zarr file.
    ds = ds.chunk({"row": data_row_chunks})
    ds.to_zarr(zarr_output_path, mode="w")
    
    #Compress the flags using numpy packbits. Groups 8 booleans into one interger.
    orig_flags_row = maintable.FLAG_ROW.values
    orig_flags = maintable.FLAG.values
    packed_flags_row = compress_bits(orig_flags_row,dim=1)
    packed_flags = compress_bits(orig_flags,dim=3)
    
    #Write the flag data to the zarr file.
    ds_flag_row = xr.Dataset({
        "FLAG_ROW":(("row"), packed_flags_row)
        },
        coords={
            "row": np.arange(packed_flags_row.shape[0])
        }
    )
    
    ds_flag_row.to_zarr(zarr_output_path, group="FLAG_ROW", mode="w")
    
    ds_flag = xr.Dataset({
        "FLAG":(("row"), packed_flags)
        },
        coords={
            "row": np.arange(packed_flags.shape[0])
        }
    )
    
    ds_flag.to_zarr(zarr_output_path, group="FLAG", mode="w")

    ds_spw = xr.Dataset(
        {
            "CHAN_WIDTH":(("row","chan"),spw_table.CHAN_WIDTH.values),
            "CHAN_FREQ":(("row","chan"),spw_table.CHAN_FREQ.values),
            "EFFECTIVE_BW":(("row","chan"),spw_table.EFFECTIVE_BW.values),
            "RESOLUTION":(("row","chan"),spw_table.RESOLUTION.values),
            "NUM_CHAN":(("row"),spw_table.NUM_CHAN.values),
            "REF_FREQUENCY":(("row"),spw_table.REF_FREQUENCY.values),
            "MEAS_FREQ_REF":(("row"),spw_table.MEAS_FREQ_REF.values),
            "TOTAL_BANDWIDTH":(("row"),spw_table.TOTAL_BANDWIDTH.values),
            "FLAG_ROW":(("row"),spw_table.FLAG_ROW.values)
        },
        coords={
            "row": np.arange((spw_table.CHAN_WIDTH.shape[0])),
            "chan": np.arange((spw_table.CHAN_WIDTH.shape[1]))
        }
    )
    
    ds_spw.to_zarr(zarr_output_path, group="SPECTRAL_WINDOW", mode="w")
    
    
    ds_pol = xr.Dataset(
        {
            "CORR_TYPE":(("row","corr"),pol_table.CORR_TYPE.values),
            "CORR_PRODUCT":(("row","corr","corrprod_idx"),pol_table.CORR_PRODUCT.values),
            "NUM_CORR":(("row"),pol_table.NUM_CORR.values)
        },
        coords={
            "row": np.arange((pol_table.CORR_TYPE.shape[0])),
            "corr": np.arange((pol_table.CORR_TYPE.shape[1])),
            "corrprod_idx": np.arange((pol_table.CORR_PRODUCT.shape[2]))
        }
    )
    
    ds_pol.to_zarr(zarr_output_path, group="POLARIZATION", mode="w") 
    
    ds_fld = xr.Dataset(
        {
           "PHASE_DIR":(("row", "field-poly", "field-dir"),fld_table.PHASE_DIR.values),
           "DELAY_DIR":(("row", "field-poly", "field-dir"),fld_table.DELAY_DIR.values),
           "REFERENCE_DIR":(("row", "field-poly", "field-dir"),fld_table.REFERENCE_DIR.values)
        },
        coords={
            "row": np.arange((fld_table.PHASE_DIR.shape[0])),
            "field-poly": np.arange((fld_table.PHASE_DIR.shape[1])),
            "field-dir": np.arange((fld_table.PHASE_DIR.shape[2]))
        }
    )
    
    ds_fld.to_zarr(zarr_output_path, group="FIELD", mode="w")
    
    
    ds_pointing = xr.Dataset(
        {
          "TARGET":(("row", "point-poly", "radec"),pntng_table.TARGET.values),
          "TIME":(("row"),pntng_table.TIME.values),
          "INTERVAL":(("row"),pntng_table.INTERVAL.values),
          "TRACKING":(("row"),pntng_table.TRACKING.values),
          "DIRECTION":(("row", "point-poly", "radec"),pntng_table.DIRECTION.values)
        },
        coords={
            "row": np.arange((pntng_table.TARGET.shape[0])),
            "point-poly": np.arange((pntng_table.TARGET.shape[1])),
            "radec": np.arange((pntng_table.TARGET.shape[2]))
        }
    )
    
    ds_pointing = ds_pointing.chunk({"row": pointing_chunks})
    
    ds_pointing.to_zarr(zarr_output_path, group="POINTING", mode="w")

    
    ds_ant = xr.Dataset(
        {
           "DISH_DIAMETER":(("row"),ant_table.DISH_DIAMETER.values),
           "POSITION":(("row","pos_dim"),ant_table.POSITION.values),
           "MOUNT":(("row"),ant_table.MOUNT.values),
           "TYPE":(("row"),ant_table.TYPE.values),
            "NAME":(("row"),ant_table.NAME.values) 
        },
        coords={
            "row": np.arange((ant_table.DISH_DIAMETER.shape[0])),
            "pos_dim": np.arange((ant_table.POSITION.shape[1]))
        }
    )
    
    ds_ant.to_zarr(zarr_output_path, group="ANTENNA", mode="w")
    
    #If decorrelation is provided, compute the required singular values for each baselines
    #based on the desired signal loss.
    if decorrelation is not None:
        
        #Go through all the baselines and correlations
        for bli in vis_data:
            ant1name = vis_data[bli]["ant1name"]
            ant2name = vis_data[bli]["ant2name"]
            baseline_key = f"{ant1name}-{ant2name}"
            
            for corr in correlation.split(','):
                
                U, singvals, WT = vis_data[bli][corr].data
                sum_total = da.sum(singvals**2).compute()
                threshold = (decorrelation)**2 * sum_total
                cumulative = da.cumsum(singvals**2).compute()
                
                n = np.argmax(cumulative >= threshold) + 1  #+1 to convert index to count
                if n == 0:
                    n = len(singvals)

                log.info(f"Baseline: {baseline_key}, n: {n}")
                m, n_orig = vis_data[bli][corr].shape
                baseline_filter = vis_data[bli]["baseline_filter"]
   
                ds_decomp = xr.Dataset(
                {
                "U": (("time", "mode"), U[:, :n].compute()),  
                "S": (("mode",), singvals[:n].compute()),
                "WT": (("mode", "channel"), WT[:n, :].compute()),
                },
                coords={
                "time": np.arange(U.shape[0]),
                "mode": np.arange(singvals.shape[0])[:n],
                "channel": np.arange(WT.shape[1]),
                },
                attrs={
                "baseline_filter": baseline_filter.tolist(),
                },
                )
                
                group_path = f"DATA/{baseline_key}/{corr}"
                ds_decomp.to_zarr(zarr_output_path, group=group_path, mode="w")
                
    
    #If compressionrank is given, use it uniformly across all the baselines.
    elif compressionrank is not None:
    
        for bli in vis_data:
            ant1name = vis_data[bli]["ant1name"]
            ant2name = vis_data[bli]["ant2name"]
            baseline_key = f"{ant1name}-{ant2name}"
             
            for corr in correlation.split(','):
                U, singvals, WT = vis_data[bli][corr].data
                m, n = vis_data[bli][corr].shape
                baseline_filter = vis_data[bli]["baseline_filter"]
                
                ds_decomp = xr.Dataset(
                {
                "U": (("time", "mode"), U[:, :compressionrank].compute()), 
                "S": (("mode",), singvals[:compressionrank].compute()),
                "WT": (("mode", "channel"), WT[:compressionrank, :].compute()),
                },
                coords={
                "time": np.arange(U.shape[0]),
                "mode": np.arange(singvals.shape[0])[:compressionrank],
                "channel": np.arange(WT.shape[1]),
                },
                attrs={
                "baseline_filter": baseline_filter.tolist(),
                "corr": corr,
                },
                )
                
                group_path = f"DATA/{baseline_key}/{corr}"
                
                ds_decomp.to_zarr(zarr_output_path, group=group_path, mode="w")
    
        log.info(f"Data successfully stored at {zarr_output_path}")
     

def compress_bits(data_array,axis=None,dim=1):
    """"
    Compress data (flags) using numpy packbits.
    
    Params
    ----
    data_array: array
    - The data to be compressed.
    
    axis: int
    - The axis along which to compress the data.
    
    dim: int
    - The dimensions of the data.
    
    Returns
    ----
    
    The compressed data that is 8 times smaller in size than the original.
    """
    #If the data has one dimension, its easy.
    if dim==1:
        #If the data is not divisible by 8, padd it at the end with booleans so that it is divisible.
        if axis==None and data_array.shape[0] % 8 != 0:
            total_length = data_array.shape[0]
            pad_size = (8 - total_length % 8) % 8
            padded_data = np.pad(data_array, ((0, pad_size)), mode='constant', constant_values=False)
            packed_data = np.packbits(padded_data,axis=None)
            
        #if the data is divisible by 8, then its easy.
        elif axis==None and data_array.shape[0] % 8 == 0:
            packed_data = np.packbits(data_array,axis=None)
            
    #If the data has more than one dimension, flatten the array.
    elif dim>1:
        
        # if the flattened array is not divisibble by 8, then pad its end.
        if axis==None and data_array.size % 8 != 0:
            total_length = data_array.size
            pad_size = (8 - total_length % 8) % 8
            padded_data = np.pad(data_array, ((0, pad_size)), mode='constant', constant_values=False)
            packed_data = np.packbits(padded_data,axis=None)

        #if the flattenen array is divisible by 8, its easy.
        elif axis==None and data_array.size % 8 ==0:
            packed_data = np.packbits(data_array,axis=None)
        
    return packed_data

    
def compress_quantized(weight_array):
     
    quantized = weight_array.astype(np.float16)
    
    return quantized


import logging
import os
from importlib import metadata

from omegaconf import OmegaConf

__version__ = metadata.version(__package__)

PCKGDIR = os.path.dirname(os.path.abspath(__file__))

SCHEMADIR = os.path.join(__path__[0], "schemas")


def get_logger(name, level="DEBUG"):

    if isinstance(level, str):
        level = getattr(logging, level, 10)

    format_string = '%(asctime)s-%(name)s-%(levelname)-8s| %(message)s'
    # set up logging to file - see previous section for more details
    logging.basicConfig(level=logging.DEBUG,
                        format=format_string,
                        datefmt='%m:%d %H:%M:%S')

    return logging.getLogger(name)


LOG = get_logger("VISCO")

BIN = OmegaConf.create({"archival": "archival",
                        "decompression": "decompression"
                        })


#Lifted from simms 3.0
class ObjDict(object):
    def __init__(self, items):
        """
        Converts a dictionary into an object. 

        """
        # First give this objects all the attributes of the input dicttionary
        for item in dir(dict):
            if not item.startswith("__"):
                setattr(self, item, getattr(items, item, None))
        # Now set the dictionary values as attributes
        self.__dict__.update(items)